@@ -3,6 +3,10 @@
 # Make sure we propegate exit codes, for kokoro.
 set -e
 
+if [ -z "$PYTHON" ]; then
+  PYTHON="python3"
+fi
+
 # Set up temp dir containing atheris and cd into it.
 SRC_DIR="$1"
 if [ -z "$SRC_DIR" ]; then
@@ -14,10 +18,10 @@ cp -r . "${TMP_DIR?}"
 cd "${TMP_DIR?}"
 
 # Set up virtual env
-python3 -m virtualenv .
+"$PYTHON" -m virtualenv .
 source bin/activate
 python -m pip install .
 python -m pip install PyInstaller
 
-cd src && python3 -m unittest discover . -p '*_test.py'
+cd src && python -m unittest discover . -p '*_test.py'
 rm -rf "$TMP_DIR"
